// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    enum_::{Enum, EnumType},
    expr::Flow,
    int::UInt,
    sim::{
        time::{SimDuration, SimInstant},
        TraceArray, TraceAsyncReset, TraceBool, TraceBundle, TraceClock, TraceDecl,
        TraceEnumDiscriminant, TraceEnumWithFields, TraceFieldlessEnum, TraceInstance,
        TraceMemPort, TraceModule, TraceModuleIO, TraceReg, TraceSInt, TraceScalar, TraceScalarId,
        TraceScope, TraceSyncReset, TraceUInt, TraceWire, TraceWriter, TraceWriterDecls,
    },
};
use bitvec::slice::BitSlice;
use std::{
    fmt::{self, Display},
    io::{self, Write},
    mem,
};

pub struct VcdWriterDecls<W: io::Write + 'static> {
    writer: W,
    timescale: SimDuration,
}

impl<W: io::Write + 'static> VcdWriterDecls<W> {
    pub fn new(writer: W) -> Self {
        Self {
            writer,
            timescale: SimDuration::from_picos(1),
        }
    }
    pub fn timescale(&self) -> SimDuration {
        self.timescale
    }
    #[track_caller]
    pub fn with_timescale(mut self, timescale: SimDuration) -> Self {
        // check timescale validity
        vcd_timescale(timescale);
        self.timescale = timescale;
        self
    }
}

#[track_caller]
const fn vcd_timescale(timescale: SimDuration) -> &'static str {
    if !timescale.is_power_of_ten() {
        panic!("VCD timescale must be a power of 10");
    }
    macro_rules! timescales {
        ($($const_name:ident = ($dur:expr, $text:literal),)*) => {
            $(const $const_name: SimDuration = $dur;)*
            match timescale {
                $($const_name => $text,)*
                _ => panic!("VCD timescale is too big"),
            }
        };
    }
    timescales! {
        TIMESCALE_1_AS = (SimDuration::from_attos(1), "1 as"),
        TIMESCALE_10_AS = (SimDuration::from_attos(10), "10 as"),
        TIMESCALE_100_AS = (SimDuration::from_attos(100), "100 as"),
        TIMESCALE_1_FS = (SimDuration::from_femtos(1), "1 fs"),
        TIMESCALE_10_FS = (SimDuration::from_femtos(10), "10 fs"),
        TIMESCALE_100_FS = (SimDuration::from_femtos(100), "100 fs"),
        TIMESCALE_1_PS = (SimDuration::from_picos(1), "1 ps"),
        TIMESCALE_10_PS = (SimDuration::from_picos(10), "10 ps"),
        TIMESCALE_100_PS = (SimDuration::from_picos(100), "100 ps"),
        TIMESCALE_1_NS = (SimDuration::from_nanos(1), "1 ns"),
        TIMESCALE_10_NS = (SimDuration::from_nanos(10), "10 ns"),
        TIMESCALE_100_NS = (SimDuration::from_nanos(100), "100 ns"),
        TIMESCALE_1_US = (SimDuration::from_micros(1), "1 us"),
        TIMESCALE_10_US = (SimDuration::from_micros(10), "10 us"),
        TIMESCALE_100_US = (SimDuration::from_micros(100), "100 us"),
        TIMESCALE_1_MS = (SimDuration::from_millis(1), "1 ms"),
        TIMESCALE_10_MS = (SimDuration::from_millis(10), "10 ms"),
        TIMESCALE_100_MS = (SimDuration::from_millis(100), "100 ms"),
        TIMESCALE_1_S = (SimDuration::from_secs(1), "1 s"),
        TIMESCALE_10_S = (SimDuration::from_secs(10), "10 s"),
        TIMESCALE_100_S = (SimDuration::from_secs(100), "100 s"),
        TIMESCALE_1000_S = (SimDuration::from_secs(1000), "1000 s"),
    }
}

impl<W: io::Write> fmt::Debug for VcdWriterDecls<W> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            writer: _,
            timescale,
        } = self;
        f.debug_struct("VcdWriterDecls")
            .field("timescale", timescale)
            .finish_non_exhaustive()
    }
}

fn write_vcd_scope<W: io::Write, R>(
    writer: &mut W,
    scope_type: &str,
    scope_name: &str,
    f: impl FnOnce(&mut W) -> io::Result<R>,
) -> io::Result<R> {
    writeln!(writer, "$scope {scope_type} {scope_name} $end")?;
    let retval = f(writer)?;
    writeln!(writer, "$upscope $end")?;
    Ok(retval)
}

macro_rules! trait_arg {
    (
        trait $Arg:ident {
            $(
                fn $fn:ident(self) -> $ty:ty;
            )*
        }
    ) => {
        trait $Arg: Sized {
            $(fn $fn(self) -> $ty {
                unreachable!()
            })*
        }

        $(
            impl $Arg for $ty {
                fn $fn(self) -> $ty {
                    self
                }
            }
        )*
    };
}

trait_arg! {
    trait Arg {
        fn module(self) -> ArgModule;
        fn module_body(self) -> ArgModuleBody;
        fn in_type(self) -> ArgInType;
    }
}

struct ArgModule {}

struct ArgModuleBody {}

#[derive(Clone, Copy)]
struct ArgInType {
    source_var_type: &'static str,
    sink_var_type: &'static str,
    duplex_var_type: &'static str,
}

trait WriteTrace: Copy {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()>;
}

impl WriteTrace for TraceDecl {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        match self {
            Self::Scope(v) => v.write_trace(writer, arg),
            Self::Scalar(v) => v.write_trace(writer, arg),
        }
    }
}

impl WriteTrace for TraceScalar {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        match self {
            Self::UInt(v) => v.write_trace(writer, arg),
            Self::SInt(v) => v.write_trace(writer, arg),
            Self::Bool(v) => v.write_trace(writer, arg),
            Self::FieldlessEnum(v) => v.write_trace(writer, arg),
            Self::EnumDiscriminant(v) => v.write_trace(writer, arg),
            Self::Clock(v) => v.write_trace(writer, arg),
            Self::SyncReset(v) => v.write_trace(writer, arg),
            Self::AsyncReset(v) => v.write_trace(writer, arg),
        }
    }
}

fn write_scalar_id<W: io::Write>(writer: &mut W, id: TraceScalarId) -> io::Result<()> {
    let min_char = b'!';
    let max_char = b'~';
    let base = (max_char - min_char + 1) as usize;
    let mut id = id.as_usize();
    loop {
        let digit = (id % base) as u8 + min_char;
        id /= base;
        writer.write_all(&[digit])?;
        if id == 0 {
            break;
        }
    }
    Ok(())
}

fn write_escaped<W: io::Write>(writer: &mut W, value: impl Display) -> io::Result<()> {
    // escaping rules from function GTKWave uses to decode VCD strings:
    // https://github.com/gtkwave/gtkwave/blob/491f24d7e8619cfc1fcc65704ee5c967d1083c18/lib/libfst/fstapi.c#L7090
    struct Wrapper<W>(W);
    impl<W: io::Write> io::Write for Wrapper<W> {
        fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
            if buf.is_empty() {
                return self.0.write(buf);
            }
            let mut retval = 0;
            for &byte in buf {
                match byte {
                    b'\\' | b'\'' | b'"' | b'?' => self.0.write_all(&[b'\\', byte])?,
                    b'\n' => self.0.write_all(br"\n")?,
                    b'\r' => self.0.write_all(br"\r")?,
                    b'\t' => self.0.write_all(br"\t")?,
                    0x7 => self.0.write_all(br"\a")?,
                    0x8 => self.0.write_all(br"\b")?,
                    0xC => self.0.write_all(br"\f")?,
                    0xB => self.0.write_all(br"\v")?,
                    _ => {
                        if byte.is_ascii_graphic() {
                            self.0.write_all(&[byte])?;
                        } else {
                            write!(self.0, r"\x{byte:02x}")?;
                        }
                    }
                }
                retval += 1;
            }
            Ok(retval)
        }

        fn flush(&mut self) -> io::Result<()> {
            self.0.flush()
        }
    }
    write!(Wrapper(writer), "{value}")
}

fn write_vcd_var<W: io::Write>(
    writer: &mut W,
    var_type: &str,
    size: usize,
    id: TraceScalarId,
    name: &str,
) -> io::Result<()> {
    write!(writer, "$var {var_type} {size} ")?;
    write_scalar_id(writer, id)?;
    writeln!(writer, " {name} $end")
}

impl WriteTrace for TraceUInt {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let ArgInType {
            source_var_type,
            sink_var_type,
            duplex_var_type,
        } = arg.in_type();
        let Self { id, name, ty, flow } = self;
        let var_type = match flow {
            Flow::Source => source_var_type,
            Flow::Sink => sink_var_type,
            Flow::Duplex => duplex_var_type,
        };
        if ty.width() == 0 {
            write_vcd_var(writer, "string", ty.width(), id, &name)
        } else {
            write_vcd_var(writer, var_type, ty.width(), id, &name)
        }
    }
}

impl WriteTrace for TraceSInt {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let Self { id, name, ty, flow } = self;
        TraceUInt {
            id,
            name,
            ty: UInt::new_dyn(ty.width()),
            flow,
        }
        .write_trace(writer, arg)
    }
}

impl WriteTrace for TraceBool {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let Self { id, name, flow } = self;
        TraceUInt {
            id,
            name,
            flow,
            ty: UInt::new_dyn(1),
        }
        .write_trace(writer, arg)
    }
}

impl WriteTrace for TraceFieldlessEnum {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let Self { id, name, ty, flow } = self;
        TraceEnumDiscriminant { id, name, ty, flow }.write_trace(writer, arg)
    }
}

impl WriteTrace for TraceEnumDiscriminant {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let ArgInType {
            source_var_type: _,
            sink_var_type: _,
            duplex_var_type: _,
        } = arg.in_type();
        let Self {
            id,
            name,
            ty: _,
            flow: _,
        } = self;
        write_vcd_var(writer, "string", 1, id, &name)
    }
}

impl WriteTrace for TraceClock {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let Self { id, name, flow } = self;
        TraceBool { id, name, flow }.write_trace(writer, arg)
    }
}

impl WriteTrace for TraceSyncReset {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let Self { id, name, flow } = self;
        TraceBool { id, name, flow }.write_trace(writer, arg)
    }
}

impl WriteTrace for TraceAsyncReset {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let Self { id, name, flow } = self;
        TraceBool { id, name, flow }.write_trace(writer, arg)
    }
}

impl WriteTrace for TraceScope {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        match self {
            Self::Module(v) => v.write_trace(writer, arg),
            Self::Instance(v) => v.write_trace(writer, arg),
            Self::MemPort(v) => v.write_trace(writer, arg),
            Self::Wire(v) => v.write_trace(writer, arg),
            Self::Reg(v) => v.write_trace(writer, arg),
            Self::ModuleIO(v) => v.write_trace(writer, arg),
            Self::Bundle(v) => v.write_trace(writer, arg),
            Self::Array(v) => v.write_trace(writer, arg),
            Self::EnumWithFields(v) => v.write_trace(writer, arg),
        }
    }
}

impl WriteTrace for TraceModule {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let ArgModule {} = arg.module();
        let Self { name, children } = self;
        write_vcd_scope(writer, "module", &name, |writer| {
            for child in children {
                child.write_trace(writer, ArgModuleBody {})?;
            }
            Ok(())
        })
    }
}

impl WriteTrace for TraceInstance {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let ArgModuleBody {} = arg.module_body();
        let Self {
            name: _,
            instance_io,
            module,
            ty: _,
        } = self;
        instance_io.write_trace(
            writer,
            ArgInType {
                source_var_type: "wire",
                sink_var_type: "wire",
                duplex_var_type: "wire",
            },
        )?;
        module.write_trace(writer, ArgModule {})
    }
}

impl WriteTrace for TraceMemPort {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        todo!()
    }
}

impl WriteTrace for TraceWire {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let ArgModuleBody {} = arg.module_body();
        let Self {
            name: _,
            child,
            ty: _,
        } = self;
        child.write_trace(
            writer,
            ArgInType {
                source_var_type: "wire",
                sink_var_type: "wire",
                duplex_var_type: "wire",
            },
        )
    }
}

impl WriteTrace for TraceReg {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let ArgModuleBody {} = arg.module_body();
        let Self {
            name: _,
            child,
            ty: _,
        } = self;
        child.write_trace(
            writer,
            ArgInType {
                source_var_type: "reg",
                sink_var_type: "reg",
                duplex_var_type: "reg",
            },
        )
    }
}

impl WriteTrace for TraceModuleIO {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let ArgModuleBody {} = arg.module_body();
        let Self {
            name: _,
            child,
            ty: _,
            flow: _,
        } = self;
        child.write_trace(
            writer,
            ArgInType {
                source_var_type: "wire",
                sink_var_type: "wire",
                duplex_var_type: "wire",
            },
        )
    }
}

impl WriteTrace for TraceBundle {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let arg = arg.in_type();
        let Self {
            name,
            fields,
            ty: _,
            flow: _,
        } = self;
        write_vcd_scope(writer, "struct", &name, |writer| {
            for field in fields {
                field.write_trace(writer, arg)?;
            }
            Ok(())
        })
    }
}

impl WriteTrace for TraceArray {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let arg = arg.in_type();
        let Self {
            name,
            elements,
            ty: _,
            flow: _,
        } = self;
        write_vcd_scope(writer, "struct", &name, |writer| {
            for element in elements {
                element.write_trace(writer, arg)?;
            }
            Ok(())
        })
    }
}

impl WriteTrace for TraceEnumWithFields {
    fn write_trace<W: io::Write, A: Arg>(self, writer: &mut W, arg: A) -> io::Result<()> {
        let arg = arg.in_type();
        let Self {
            name,
            discriminant,
            non_empty_fields,
            ty: _,
            flow: _,
        } = self;
        write_vcd_scope(writer, "struct", &name, |writer| {
            discriminant.write_trace(writer, arg)?;
            for field in non_empty_fields {
                field.write_trace(writer, arg)?;
            }
            Ok(())
        })
    }
}

impl<W: io::Write> TraceWriterDecls for VcdWriterDecls<W> {
    type Error = io::Error;
    type TraceWriter = VcdWriter<W>;

    fn write_decls(self, module: TraceModule) -> Result<Self::TraceWriter, Self::Error> {
        let Self {
            mut writer,
            timescale,
        } = self;
        writeln!(writer, "$timescale {} $end", vcd_timescale(timescale))?;
        module.write_trace(&mut writer, ArgModule {})?;
        writeln!(writer, "$enddefinitions $end")?;
        writeln!(writer, "$dumpvars")?;
        Ok(VcdWriter {
            writer,
            finished_init: false,
            timescale,
        })
    }
}

pub struct VcdWriter<W: io::Write + 'static> {
    writer: W,
    finished_init: bool,
    timescale: SimDuration,
}

impl<W: io::Write + 'static> VcdWriter<W> {
    pub fn timescale(&self) -> SimDuration {
        self.timescale
    }
    fn write_string_value_change(
        &mut self,
        value: impl Display,
        id: TraceScalarId,
    ) -> io::Result<()> {
        self.writer.write_all(b"s")?;
        write_escaped(&mut self.writer, value)?;
        self.writer.write_all(b" ")?;
        write_scalar_id(&mut self.writer, id)?;
        self.writer.write_all(b"\n")
    }
}

impl<W: io::Write> TraceWriter for VcdWriter<W> {
    type Error = io::Error;

    fn set_signal_uint(&mut self, id: TraceScalarId, value: &BitSlice) -> Result<(), Self::Error> {
        match value.len() {
            0 => self.writer.write_all(b"s0 ")?,
            1 => self.writer.write_all(if value[0] { b"1" } else { b"0" })?,
            _ => {
                self.writer.write_all(b"b")?;
                let mut any_ones = false;
                for bit in value.iter().rev() {
                    if *bit {
                        any_ones = true;
                        self.writer.write_all(b"1")?;
                    } else if any_ones {
                        self.writer.write_all(b"0")?;
                    }
                }
                if !any_ones {
                    self.writer.write_all(b"0")?;
                }
                self.writer.write_all(b" ")?;
            }
        }
        write_scalar_id(&mut self.writer, id)?;
        self.writer.write_all(b"\n")
    }

    fn set_signal_sint(&mut self, id: TraceScalarId, value: &BitSlice) -> Result<(), Self::Error> {
        self.set_signal_uint(id, value)
    }

    fn finish_init(&mut self) -> Result<(), Self::Error> {
        if mem::replace(&mut self.finished_init, true) {
            return Ok(());
        }
        writeln!(self.writer, "$end")
    }

    fn change_time_to(&mut self, instant: SimInstant) -> Result<(), Self::Error> {
        assert!(self.finished_init);
        let mut instant_attos = (instant - SimInstant::START).as_attos();
        instant_attos += self.timescale.as_attos() / 2;
        let timestamp = instant_attos / self.timescale.as_attos();
        writeln!(self.writer, "#{timestamp}")
    }

    fn flush(&mut self) -> Result<(), Self::Error> {
        self.writer.flush()
    }

    fn close(mut self) -> Result<(), Self::Error> {
        self.writer.flush()
    }

    fn set_signal_enum_discriminant(
        &mut self,
        id: TraceScalarId,
        variant_index: usize,
        ty: Enum,
    ) -> Result<(), Self::Error> {
        self.write_string_value_change(
            format_args!(
                "{} ({variant_index})",
                ty.variants()
                    .get(variant_index)
                    .map(|v| &*v.name)
                    .unwrap_or("<invalid>"),
            ),
            id,
        )
    }
}

impl<W: io::Write> fmt::Debug for VcdWriter<W> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            writer: _,
            finished_init,
            timescale,
        } = self;
        f.debug_struct("VcdWriter")
            .field("finished_init", finished_init)
            .field("timescale", timescale)
            .finish_non_exhaustive()
    }
}
